# pygtktest.py
# 
# Very short test file to see if the installation of PyGTK was successful.
#
# Author       : Chris X. Weichenberger
# Created      : unk
# License
# Twilight is a GUI for displaying ligands incompatible with electron density.
#    Copyright (C) 2012  Christian X. Weichenberger, Edwin Pozharski, and
#    Bernhard Rupp.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import pygtk
pygtk.require('2.0')
import gtk
import gobject;

window = gtk.MessageDialog(buttons=gtk.BUTTONS_OK_CANCEL);
window.connect("destroy", lambda wid: gtk.main_quit());
window.connect("delete_event", lambda a1,a2:gtk.main_quit());
window.set_property("text", "You did it!")
window.run()
